import re
from typing import Dict, List, Tuple


class Parser:
    def __init__(self, log_file: str) -> None:
        """
        Initialize the Parser with a log file.

        Parameters:
            log_file: The name of the log file.
        """
        self.log_file = log_file
        self.log_data = self.read_log()

    def read_log(self) -> List[str]:
        """
        Reads the log file and returns its content.

        Returns:
            List[str]: The content of the log file as a list of strings.
        """
        with open(self.log_file, "r") as log_file:
            log_data = log_file.readlines()
        return self.split_log_to_sections(log_data)

    def split_log_to_sections(self, log_data: List[str]) -> Dict:
        """
        Splits the log file into sections based on specified markers.

        Parameters:
            log_data (List[str]): Log content read with read_log method.

        Returns:
            Dict: A dictionary with section names as keys
                and content as values.
        """
        section_pattern = r"\{\{(.+?)\}\}"
        section_name = "beginning"
        log_by_sections: Dict[str, List[str]] = {section_name: []}

        for line in log_data:
            if " ::: " in line:
                line = line.split(" ::: ")[1].strip()
                match_section_pattern = re.search(section_pattern, line)
                if match_section_pattern:
                    section_name = "_".join(line.split()[1:-1]).lower()
                    log_by_sections[section_name] = []
                else:
                    log_by_sections[section_name].append(line)
            else:
                log_by_sections["final"] = line

        return log_by_sections

    def parse_emerge_pretend_section(self, section_content: List[str]) -> Dict:
        """
        Function to parse the "emerge pretend" section of the log data.

        Parameters:
            section_content (List[str]): A list of strings that contains
                        the content of the "emerge pretend" section.

        Returns:
            Dict: A dictionary that contains the status of
                  the "emerge pretend" operation and the details.
        """
        if "emerge pretend was successful, updating..." in section_content:
            pretend_status = True
            pretend_details = None
            return {
                "pretend_status": pretend_status,
                "pretend_details": pretend_details,
            }
        else:
            pretend_status = False
            # function to parse errors during emerge --pretend
            pretend_details = self.parse_pretend_details(section_content)
            return {
                "pretend_status": pretend_status,
                "pretend_details": pretend_details,
            }

    def _parse_pretend_get_blocked_details(
        self, error_content: List[str]
    ) -> List[str]:
        """
        Parse details of blocked package error.

        Parameters:
            error_content (List[str]): Lines from log marked as ERROR by logger.

        Returns:
            List[str]: Parsed packages tht cause the error.
        """
        blocked_packages = []
        for line in error_content:
            match_package_pattern = re.search(r"^\((.+?)\)", line)
            if match_package_pattern:
                blocked_packages.append(match_package_pattern.group(1))

        return blocked_packages

    def _parse_pretend_get_error_type(
        self, error_content
    ) -> Tuple[str, str, List[str]]:
        """
        Parse error type and error details.

        Parameters:
            error_content (List[str]): Lines from log marked as ERROR by logger.

        Returns:
            Tuple[str, str, List[str]]: Return error type, error definition as strings
                and error details as a list of strings.
        """
        error_type = "undefined"
        error_definition = "undefined"
        error_details = "undefined"

        for line in error_content:
            if "Blocked Packages" in line:
                error_type = "Blocked Packages"
                error_details = self._parse_pretend_get_blocked_details(
                    error_content
                )

        for line in error_content:
            if line[0] == "*":
                error_definition += f"{line[2:]} "

        return (error_type, error_definition, error_details)

    def parse_pretend_details(self, section_content: List[str]) -> Dict:
        """
        Parse information about the emerge pretend from logs.

        Parameters:
            section_content (List[str]): A list where each item is
                one line of logs from a section.

        Returns:
            List[Dict]: A list of dictionaries where each item is
                a named dictionary containing useful information for the report.
        """
        error_index = section_content.index(
            "emerge pretend has failed, exiting"
        )
        error_content = [
            line for line in section_content[error_index + 1 :] if line
        ]

        error_type, _, error_details = self._parse_pretend_get_error_type(
            error_content
        )
        pretend_details = {
            "error_type": error_type,
            "error_details": error_details,
        }
        return pretend_details

    def parse_update_system_section(self, section_content: List[str]) -> Dict:
        """
        Function to parse the "update system" section of the log data.

        Parameters:
            section_content (List[str]): A list of strings that contains
                    the content of the "update system" section.

        Returns:
            Dict: A dictionary that contains the status
                  of the system update and the details.
        """
        update_type = section_content[1].split()[1]
        if "update was successful" in section_content:
            update_status = True
            package_list = self.parse_update_details(section_content)
            update_details = {"updated_packages": package_list}
            return {
                "update_type": update_type,
                "update_status": update_status,
                "update_details": update_details,
            }
        else:
            update_status = False
            # function to parse emerge update errors
            # update_details = parse_update_error_details(section_content, type)
            update_details = {"updated_packages": [], "errors": []}
            return {
                "update_type": update_type,
                "update_status": update_status,
                "update_details": update_details,
            }

    def _parse_update_get_packages_names(
        self, name_and_version: str
    ) -> Tuple[str, str]:
        """
        Parse package name and new version.

        Parameters:
            name_and_version (str): String containing package information.
            example:
                sys-process/procps-3.3.17-r2:0/8::gentoo

        Returns:
            Tuple: tuple containing package name and it's new version
            example:
                (sys-process/procps, 3.3.17-r2:0/8)
        """
        regex_package_info = (
            r"^(.+?)-"  # regex_package_name
            r"("
            r"(?:(?<=-)[a-z]{2,}|[0-9]+(?:\.[0-9]+)*)"  # regex_version_number
            r"(?:_p[0-9]+)?"  # regex_optional_patch
            r"(?:-r[0-9]+)?"  # regex_optional_revision
            r"(?::[0-9]+(?:/[0-9]+(\.[0-9]+)?)?)?"  # regex_optional_sub_version
            r")"
            r"(?::|$)"  # regex end
        )
        package_regex = re.compile(regex_package_info)
        match = package_regex.search(name_and_version)
        if match:
            package_name, new_version, _ = match.groups()
            return package_name, new_version
        else:
            return ("undefined", "undefined")

    def parse_update_details(self, section_content: List[str]) -> List[Dict]:
        """
        Parse information about the update from logs.

        Parameters:
            section_content (List[str]): A list where each item is
                one line of logs from a section.

        Returns:
            List[Dict]: A list of dictionaries where each item is
                a named dictionary containing useful information for the report.
        """
        ebuild_info_pattern = r"\[(.+?)\]"
        package_strings = [
            line
            for line in section_content
            if re.search(ebuild_info_pattern, line) and line != "[ ok ]"
        ]
        packages = []
        for line in package_strings:
            chunks = re.findall(r'\S+=".*?"|\[.*?\]|\S+', line)
            package_name, new_version = self._parse_update_get_packages_names(
                chunks[1]
            )
            ebuild_info = {
                package_name: {
                    "New Version": new_version,
                    "Old Version": "undefined",
                    "Update Status": "undefined",
                    "USE Flags": "undefined",
                    "Multilibs": "undefined",
                }
            }

            for chunk in chunks:
                # Match update status
                update_status = re.match(r"^(\[ebuild\s+.*\])$", chunk)
                if update_status:
                    ebuild_info[package_name][
                        "Update Status"
                    ] = update_status.group(1)

                old_version = re.search(r"\[(.*)(::gentoo)\]", chunk)
                if old_version:
                    ebuild_info[package_name][
                        "Old Version"
                    ] = old_version.group(1)

                # Match USE flags
                use_flags = re.search(r'USE="(.*?)"', chunk)
                if use_flags:
                    ebuild_info[package_name]["USE Flags"] = use_flags.group(1)

                # Match multilibs
                multilibs = re.search(r'(ABI_X86=".*?")', chunk)
                if multilibs:
                    ebuild_info[package_name]["Multilibs"] = multilibs.group(1)

            packages.append(ebuild_info)
        return packages

    def parse_disk_usage_info(self, section_content: List[str]) -> Dict:
        """
        Get disk usage information.

        Parameters:
            section_content (List[str]): A list where each item is
                one line of logs from a section.

        Returns:
            Dict: A dictionary containing statistics of disk usage.
        """
        disk_usage = {}
        split_content = section_content[1].split(" ===> ")
        split_content = split_content[1].split(", ")
        for stat in split_content:
            stat = stat.split("=")
            disk_usage[stat[0]] = stat[1]

        return disk_usage

    def extract_info_for_report(self) -> Dict:
        """
        Extract information about the update from the log file.

        Returns:
            Dict: A dictionary that contains the
                  parsed data from all sections.
        """
        info = {"disk_usage": {}}
        for section in self.log_data.keys():
            section_content = self.log_data[section]
            if section == "pretend_emerge":
                info[section] = self.parse_emerge_pretend_section(
                    section_content
                )
            elif section == "update_system":
                info[section] = self.parse_update_system_section(
                    section_content
                )
            elif "calculate_disk_usage" in section:
                info["disk_usage"][section] = self.parse_disk_usage_info(
                    section_content
                )

        return info



from .gentoo_update import add_prefixes, create_cli, main
from .shell_runner import ShellRunner
from .parser import Parser
from .reporter import Reporter
from .notifier import Notifier
from ._version import __version__



from typing import Dict, List


class Reporter:
    def __init__(self, update_info) -> None:
        self.info = update_info
        self.report = self.create_report()

    def create_failed_pretend_report(self, pretend_info: Dict) -> List[str]:
        """
        Create a report when emerge pretend fails.

        Parameters:
            pretend_info (Dict): emerge pretend information parsed by
                parse_emerge_pretend_section.

        Returns:
            List: A list of strings that comprise the failed pretend report.
        """
        report = [
            "==========> Gentoo Update Report <==========",
            "emerge pretend status: FAIL",
        ]
        pretend_details = pretend_info["pretend_details"]
        if pretend_details["error_type"] == "undefined":
            report.append("Could not identify error, please check the logs")
        else:
            report.append(f"\nError Type: {pretend_details['error_type']}")
            report.extend(self._report_blocked_packages(pretend_details))
        report.append("")
        return report

    def _report_blocked_packages(self, pretend_details: List[str]) -> List[str]:
        """
        Report on Blocked Packages error during emerge pretend.

        Parameters:
            pretend_details (List[str]): In this case it's blocked packages.

        Returns:
            List[str]: Section of the report about blocked packages.
        """
        blocked_packages_report = ["List of Blocked Packages:"]
        for package in pretend_details["error_details"]:
            blocked_packages_report.append(f"-----> {package}")

        return blocked_packages_report

    def create_failed_report(
        self, update_info: List[Dict], disk_usage_info: Dict
    ) -> List[str]:
        """
        Create a report when update failes.

        Parameters:
            update_info (List[Dict]): Update information parsed by
                self.parse_update_details.
            disk_usage_inf (Dict): Disk usage information parsed by
                self.parse_disk_usage_info.

        Returns:
            List: A list of strings that comprise the failed update report.
        """
        # do failed report processing
        return ["Your update is a failure"]

    def create_successful_report(
        self, update_info: List[Dict], disk_usage_info: Dict
    ) -> List[str]:
        """
        Create a report when update succeeds.

        Parameters:
            update_info (List[Dict]): Update information parsed by
                self.parse_update_details.
            disk_usage_inf (Dict): Disk usage information parsed by
                self.parse_disk_usage_info.

        Returns:
            List: A list of strings that comprise the successful update report.
        """
        report = [
            "==========> Gentoo Update Report <==========",
            "update status: SUCCESS",
        ]
        updated_packages = update_info["update_details"]["updated_packages"]
        if updated_packages:
            report.append(r"processed packages:")
            for package in updated_packages:
                package_name = list(package.keys())[0]
                new_version = package[package_name]["New Version"]
                old_version = package[package_name]["Old Version"]
                report.append(
                    f"--- {package_name} {old_version}->{new_version}"
                )
            report.append("")

            disk_usage_before = disk_usage_info["calculate_disk_usage_1"]
            disk_usage_after = disk_usage_info["calculate_disk_usage_2"]
            disk_usage_stats = (
                "Disk Usage Stats:\n"
                f"Free Space {disk_usage_before['Free']} => {disk_usage_after['Free']}\n"
                f"Used Space {disk_usage_before['Used']} => {disk_usage_after['Used']}\n"
                f"Used pc(%) {disk_usage_before['Percent used']} => {disk_usage_after['Percent used']}\n"
            )
            report.append(disk_usage_stats)

        return report

    def create_report(self) -> List[str]:
        """
        Create a report.

        Returns:
            List: A list of strings that comprise the update report.
        """
        info = self.info
        disk_usage_info = info["disk_usage"]
        pretend_success = info["pretend_emerge"]["pretend_status"]

        if pretend_success:
            update_info = info["update_system"]
            update_success = info["update_system"]["update_status"]
        else:
            pretend_info = info["pretend_emerge"]
            update_success = False

        if update_success:
            report = self.create_successful_report(update_info, disk_usage_info)
            return report
        elif pretend_success and not update_success:
            report = self.create_failed_report(update_info, disk_usage_info)
        else:
            report = self.create_failed_pretend_report(pretend_info)
            return report



import os
import socket
import ssl
import time
from sys import exit
from typing import List, Tuple

USE_SENDGRID = True
try:
    import sendgrid
    from sendgrid.helpers.mail import Mail, Email, To, Content
except ImportError:
    USE_SENDGRID = False


class Notifier:
    def __init__(
        self, notification_type: str, report: List, short=True
    ) -> None:
        report = report[0:2] if short else report

        if notification_type == "email":
            if USE_SENDGRID:
                self.send_report_to_mail(report)
            else:
                print("sendgrid library is not installed")
                print("it can be installed from GURU overlay:")
                print("  emerge --ask dev-python/sendgrid")
        elif notification_type == "irc":
            self.send_report_to_irc(report)
        else:
            print("Unsupported authentication methods")
            print("Currently supporting: irc")
            print("Exiting...")

    def get_irc_vars(self) -> Tuple[str, str, str]:
        """
        Get variables needed to send report to IRC chat from env.
        """
        channel = os.getenv("IRC_CHANNEL")
        botnick = os.getenv("IRC_BOT_NICKNAME")
        botpass = os.getenv("IRC_BOT_PASSWORD")
        if None not in (channel, botnick, botpass):
            return channel, botnick, botpass
        else:
            print("Undefined enviromental variable(s)")
            print(
                "Please define: IRC_CHANNEL, IRC_BOT_NICKNAME, IRC_BOT_PASSWORD"
            )
            exit(1)

    def send_report_to_irc(self, report: List[str]) -> None:
        """
        Send the update report to IRC chat.
        """
        server = "irc.libera.chat"
        port = 6697
        channel, botnick, botpass = self.get_irc_vars()
        ssl_context = ssl.create_default_context()

        irc = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        irc = ssl_context.wrap_socket(irc, server_hostname=server)
        irc.connect((server, port))
        irc.send(
            f"USER {botnick} {botnick} {botnick} {botnick}\n".encode("UTF-8")
        )
        irc.send(f"NICK {botnick}\n".encode("UTF-8"))
        irc.send(
            f"PRIVMSG NickServ :IDENTIFY {botnick} {botpass}\n".encode("UTF-8")
        )
        time.sleep(5)

        irc.send(f"JOIN {channel}\n".encode("UTF-8"))
        for line in report:
            irc.send(f"PRIVMSG {channel} :{line}\n".encode("UTF-8"))
            time.sleep(10)
        print("report sent, quitting...")

        irc.send("QUIT \n".encode("UTF-8"))
        irc.close()

    def get_mail_vars(self) -> Tuple[str, str, str]:
        """
        Get variables needed to send report to email via SendGrid from env.
        """
        api_key = os.getenv("SENDGRID_API_KEY")
        send_to = os.getenv("SENDGRID_TO")
        send_from = os.getenv("SENDGRID_FROM")
        if None not in (api_key, send_to, send_from):
            return api_key, send_to, send_from
        else:
            print("Undefined enviromental variable(s)")
            print("Please define: SENDGRID_API_KEY, SENDGRID_TO, SENDGRID_FROM")
            exit(1)

    def send_report_to_mail(self, report: List[str]) -> None:
        """
        Send the update report to email via SendGrid.
        """
        api_key, send_to, send_from = self.get_mail_vars()
        sendgrid_client = sendgrid.SendGridAPIClient(api_key=api_key)
        subject = "Gentoo Linux Update Report"

        content = Content("text/plain", "\n".join(report))
        mail = Mail(Email(send_from), To(send_to), subject, content)
        mail_json = mail.get()

        response = sendgrid_client.client.mail.send.post(request_body=mail_json)
        if response.status_code == 202:
            print("email was sent successfully!")
        else:
            print("email was not sent successfully, details:")
            print(response)


if __name__ == "__main__":
    report = [
        "==========> Gentoo Update Report <==========",
        "update status: SUCCESS",
        "processed packages:",
        "--- sys-apps/sandbox 2.32->2.37",
        "",
        "Disk Usage Stats:",
        "Free Space 35G => 35G",
        "Used Space 7.5G => 7.5G",
        "Used pc(%) 18% => 18%",
    ]
    notify = Notifier(notification_type="email", report=report, short=False)



import unittest
from unittest.mock import patch, Mock, MagicMock
import gentoo_update
from gentoo_update import create_cli, add_prefixes


class TestGentooUpdate(unittest.TestCase):
    def test_add_prefixes(self):
        input_args = "a bc"
        expected_output = "-a --bc"
        self.assertEqual(add_prefixes(input_args), expected_output)

    @patch("argparse.ArgumentParser.parse_args")
    def test_create_cli(self, mock_parse_args):
        mock_args = Mock()
        mock_parse_args.return_value = mock_args
        args = create_cli()
        self.assertEqual(args, mock_args)

    @patch("gentoo_update.gentoo_update.ShellRunner")
    @patch("argparse.ArgumentParser.parse_args")
    @patch("gentoo_update.gentoo_update.add_prefixes")
    def test_main(self, mock_add_prefixes, mock_parse_args, mock_shell_runner):
        mock_args = Mock()
        mock_args.quiet = "y"
        mock_args.update_mode = "security"
        mock_args.args = None
        mock_args.config_update_mode = "ignore"
        mock_args.daemon_restart = "n"
        mock_args.clean = "n"
        mock_args.read_logs = "n"
        mock_args.read_news = "n"
        mock_parse_args.return_value = mock_args

        mock_add_prefixes.return_value = []

        mock_shell_runner_instance = mock_shell_runner.return_value
        mock_shell_runner_instance.run_shell_script = MagicMock()

        gentoo_update.gentoo_update.main()
        mock_shell_runner_instance.run_shell_script.assert_called_once()


if __name__ == "__main__":
    unittest.main()
